#define _NO_CRT_STDIO_INLINE
#define _CRT_SECURE_NO_WARNINGS
#define _WIN32_WINNT 0x0500

#include <windows.h>
#include <Shlwapi.h>
#pragma intrinsic(memcpy,memcmp,memset)
#define elemof(x) (sizeof(x)/sizeof(*(x)))

/****************************************
 * Ersatz fr std::vector und std::map	*
 * mit fester Speicherallokation	*
 ****************************************/
template<class T,size_t N> class static_vector {
  size_t fill;
  T data[N];
public:
  typedef T* iterator;
  typedef const T* const_iterator;
  static_vector():fill(0) {}
  size_t size() const {return fill;}
  iterator begin() {return data;}
  iterator end() {return data+fill;}
  const_iterator begin() const {return data;}
  const_iterator end() const {return data+fill;}
  bool push_back(const T&el) {if (fill>=N) return false; data[fill++]=el; return true;}
  T&operator[](size_t i) {return data[i];}
  const T&operator[](size_t i) const {return data[i];}
  void erase(iterator it) {memcpy(it,it+1,(data[--fill]-it)*sizeof(T));}
  void clear() {fill=0;}
};

template<class K,class V,size_t N> class static_map {
  size_t fill;
  K keys[N];
  V vals[N];
public:
  typedef K* iterator;
  typedef const K* const_iterator;
  static_map():fill(0) {}
  size_t size() const {return fill;}
  iterator begin() {return keys;}
  iterator end() {return keys+fill;}
  const_iterator begin() const {return keys;}
  const_iterator end() const {return keys+fill;}
  const_iterator find(const K&k) const{
    const_iterator r=begin();
    for(;r!=end();r++) if (k==*r) break;
    return r;
  }
  size_t erase(const K&k) {const_iterator f(find(k)); if (f==end()) return 0; erase(f); return 1;}
  void erase(const_iterator it) {
    size_t i=it-keys;
    --fill;
    memcpy(keys+i,keys+i+1,(fill-i)*sizeof(K));
    memcpy(vals+i,vals+i+1,(fill-i)*sizeof(V));
  }
  void clear() {fill=0;}
  V&operator[](iterator it) {
    return vals[it-keys];
  }
  V&operator[](const K&k) {
    const_iterator f(find(k));
    if (f==end() && fill<N) keys[fill++]=k;
    return vals[f-keys];
  }
  void swap(static_map&other) {
    size_t*a=(size_t*)this;
    size_t*b=(size_t*)&other;
    size_t len=sizeof*this/sizeof(size_t);
    do {
      size_t t=*a; *a++=*b; *b++=t;
    } while(--len);
  }
};

/************************
 * Aufzhlungs-Objekte	*
 ************************/
struct Rect:public RECT {
  LONG width() const {return right-left;}
  LONG height() const {return bottom-top;}
  BOOL operator==(const RECT&r) const {return EqualRect(this,&r);}
  bool operator!=(const RECT&r) const {return !operator==(r);}
  bool sameSize(const Rect&r) const {return width()==r.width() && height()==r.height();}
};

struct monvec:public static_vector<Rect,8> {
  static BOOL CALLBACK EnumMonitorProc(HMONITOR,HDC,RECT*,LPARAM);
  void GetAllMonitors();
};

struct Placement:public WINDOWPLACEMENT {
  bool get(HWND);
  BOOL set(HWND w,char show=SW_SHOWNOACTIVATE) {
    showCmd=show;
    flags |= WPF_ASYNCWINDOWPLACEMENT;
    return SetWindowPlacement(w,this);
  }
  Rect&rcNormal() {return *(Rect*)&rcNormalPosition;}
  const Rect&rcNormal() const {return *(const Rect*)&rcNormalPosition;}
};

/************************
 * Statische Daten	*
 ************************/
typedef static_map<HWND,Placement,256> placemap;
static monvec DesiredMonitorPlacements;
static placemap WindowPlacements;
static placemap PendingUpdates0;
static placemap PendingUpdates1;
static bool IsUpdateScheduled;
static bool IsPaused;
static HWND DetectWindow;
static int ResumeTimerValue = 1000;
static int ScheduleTimerValue = 1000;
static HANDLE logfile;

// _____________________________________________________________________ //
//
// Helpers
// _____________________________________________________________________ //
static int clamp(int min, int max, int value) {
  if (value<min) value=min;
  if (value>max) value=max;
  return value;
}

static void _cdecl Log(const TCHAR*fmt, ...) {
  TCHAR buf[256];
  va_list args;
  va_start(args, fmt);
  DWORD len=wvnsprintf(buf,elemof(buf),fmt,args);
  va_end(args);
  OutputDebugString(buf);
  if (logfile) {
#ifdef UNICODE
    char b2[256];
    DWORD l2=WideCharToMultiByte(CP_UTF8,0,buf,len,b2,sizeof b2,0,0);
    WriteFile(logfile,b2,l2,&l2,0);
#else
    WriteFile(logfile,buf,len,&len,0);
#endif
  }
}

static TCHAR*WndText(HWND hWnd) {
  static TCHAR title[100];
  GetWindowText(hWnd,title,elemof(title));
  return title;
}

static TCHAR*PlacementText(const Placement&placement) {
  static TCHAR text[100];
  const Rect&R=placement.rcNormal();
  wnsprintf(text,elemof(text),TEXT("%d, %d (%d%d)\n"),R.left,R.top,R.width(),R.height());
  return text;
}

#ifdef _DEBUG
#define debug(x) Log x
#else
#define debug(x)
#endif

/************************
 * Monitor placement	*
 ************************/
BOOL CALLBACK monvec::EnumMonitorProc(HMONITOR hMonitor,HDC,RECT*rect,LPARAM lParam) {
  monvec*self=(monvec*)lParam;
  self->push_back(*(Rect*)rect);
  return TRUE;
}

void monvec::GetAllMonitors() {
  clear();
  EnumDisplayMonitors(0,0,EnumMonitorProc,(LPARAM)this);
}

static bool IsDesiredMonitorLayout() {
  monvec monitors;
  monitors.GetAllMonitors();
  if (monitors.size() != DesiredMonitorPlacements.size()) return false;
	// This assumes deterministic order which might not work on every system
  for (size_t i = 0; i < monitors.size(); i++) {	// Only care about the size
    if (!monitors[i].sameSize(DesiredMonitorPlacements[i])) return false;
  }
  return true;
}

static bool IsIdenticalMonitorLayout() {
  monvec monitors;
  monitors.GetAllMonitors();
  if (monitors.size() != DesiredMonitorPlacements.size()) return false;
	// This assumes deterministic order which might not work on every system
  for (size_t i = 0; i < monitors.size(); i++) {
    if (monitors[i]!=DesiredMonitorPlacements[i]) return false;
  }
  return true;
}

/************************
 * Window placement	*
 ************************/
static bool IsApplicationWindow(HWND hWnd) {
  return IsWindowVisible(hWnd) && !GetParent(hWnd);
}

bool Placement::get(HWND hWnd) {
  length = sizeof(WINDOWPLACEMENT);
  if (!GetWindowPlacement(hWnd,this)) return false;
  if (showCmd == SW_SHOWNORMAL) {
	// If the window is "docked" the normalposition contains the last non-docked position.
	// That is not where we want to restore it, so extract the current rect with GetWindowRect
	// and translate the position to client coordinates
    GetWindowRect(hWnd,&rcNormalPosition);
	// From MSDN WINDOWPLACEMENT reference:
	// "If the window is a top-level window that does not have the WS_EX_TOOLWINDOW window style,
	// then the coordinates represented by the following members are in workspace coordinates"
	// "Otherwise, these members are in screen coordinates"
    DWORD exStyle = GetWindowLong(hWnd, GWL_EXSTYLE);
    if (!(exStyle&WS_EX_TOOLWINDOW)) {
      MONITORINFO monitorInfo;
      monitorInfo.cbSize = sizeof(MONITORINFO);
      HMONITOR hMonitor = MonitorFromWindow(hWnd, MONITOR_DEFAULTTONEAREST);
      GetMonitorInfo(hMonitor, &monitorInfo);
      int dx = monitorInfo.rcMonitor.left - monitorInfo.rcWork.left;
      int dy = monitorInfo.rcMonitor.top - monitorInfo.rcWork.top;
      OffsetRect(&rcNormalPosition,dx,dy);
    }
  }
  return true;
}

static void AddWindow(HWND hWnd) {
  Placement placement;
  if (placement.get(hWnd)) {
    WindowPlacements[hWnd] = placement;
    debug((TEXT("Adding %s\n\tat %s\n"),WndText(hWnd),PlacementText(placement)));
  }
}

static void RemoveWindow(HWND hWnd) {
  if (WindowPlacements.erase(hWnd)) {
    debug((TEXT("Removing %s\n"),WndText(hWnd)));
    PendingUpdates0.erase(hWnd);
    PendingUpdates1.erase(hWnd);
  }
}

static bool HasWindow(HWND hWnd) {
  return WindowPlacements.find(hWnd) != WindowPlacements.end();
}

static BOOL CALLBACK EnumWindowsProc(HWND w,LPARAM) {
  if (IsApplicationWindow(w)) AddWindow(w);
  return TRUE;
}

static void LoadAllWindowPlacements() {
  WindowPlacements.clear();
  EnumWindows(EnumWindowsProc,0);
}

static void CALLBACK UpdateWindowPlacements(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime) {
  if (!IsUpdateScheduled) {
    KillTimer(hwnd, idEvent);
    return;
  }
#ifdef _DEBUG
  if (PendingUpdates1.size()) debug((TEXT("Updating WindowPlacements\n")));
#endif
  for (placemap::iterator it = PendingUpdates1.begin(); it != PendingUpdates1.end(); ++it) {
    WindowPlacements[*it] = PendingUpdates1[it];
  }
  PendingUpdates1.clear();
  PendingUpdates0.swap(PendingUpdates1);
  if (PendingUpdates1.size()) {
    debug((TEXT("Scheduling update\n")));
    SetTimer(0,0,ScheduleTimerValue,UpdateWindowPlacements);
  }else{
    KillTimer(hwnd,idEvent);
    IsUpdateScheduled=false;
  }
}

static void ScheduleWindowPlacementUpdate(HWND hWnd) {
  Placement placement;
  if (!placement.get(hWnd)) return;
  PendingUpdates0[hWnd] = placement;
  if (!IsUpdateScheduled) {
    debug((TEXT("Scheduling update\n")));
    IsUpdateScheduled=true;
    SetTimer(0,0,ScheduleTimerValue,UpdateWindowPlacements);
  }
}

static void RestoreWindowPlacements() {
  for (placemap::iterator it = WindowPlacements.begin(); it != WindowPlacements.end(); ++it) {
    Placement&placement = WindowPlacements[it];
	// The only thing that change when a monitor is connected or disconnected is the
	// window's "normal position". Everything else remain the same (Z order,
	// maximized/minimized status, focus, etc)"
	// Since the maximized status cannot be catched with SetWinEventHook we have
	// to extract the placement again
    Placement current;
    if (!current.get(*it)) continue;
	// No need to restore it if it's already in its correct position
    if (!memcmp(&current, &placement, sizeof(current)))	continue;

    if (GetWindowTextLength(*it)) {
      debug((TEXT("%s\n"), WndText(*it)));
      debug((TEXT("\tRestore %s to %s\n"), current.showCmd == SW_SHOWMINIMIZED ? TEXT("minimized") : (current.showCmd == SW_SHOWMAXIMIZED ? TEXT("maximized") : TEXT("")), PlacementText(placement)));
    }
    if (current.showCmd == SW_SHOWMINIMIZED) {
      placement.set(*it,SW_SHOWMINNOACTIVE);	// Restore its minimized position
    }else if (current.showCmd == SW_SHOWMAXIMIZED) {
	// The window was maximized
	// In order to restore the window on the correct display we have to move it to its normal position first, and then maximize it.
	// If we only maximize it (without moving it first) it will be maximized on the display it's currently on.
	// Before we maximize it we have to show it in its normal state. Otherwise it will retain its size on the new monitor which
	// will be incorrect if the new monitor has a different resolution
      ShowWindowAsync(*it,SW_SHOWNOACTIVATE);	// Restore
      placement.set(*it);			// Move
      ShowWindowAsync(*it,SW_SHOWMAXIMIZED);	// Maximize
    }else{
      placement.set(*it);	// Restore its normal position
    }
  }
}

static void PauseWindowTracking() {
  if (IsPaused) return;
  debug((TEXT(" =========== PAUSE WINDOW TRACKING ===========\n")));
  PendingUpdates0.clear();
  PendingUpdates1.clear();
  IsPaused = true;
}

static void CALLBACK ResumeWindowTracking(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime) {
  KillTimer(hwnd, idEvent);
  if (!IsPaused) return;
	// The layout could've changed since the call to ScheduleResumeWindowTracking()
  if (!IsDesiredMonitorLayout()) return;
  debug((TEXT(" =========== RESUME WINDOW TRACKING ===========\n")));
  IsPaused = false;
  RestoreWindowPlacements();
}

static void ScheduleResumeWindowTracking() {
  static UINT_PTR ResumeTimer;
  ResumeTimer = SetTimer(0, ResumeTimer, ResumeTimerValue, ResumeWindowTracking);
}

/***********************************************
 * Dummy window for broadcast message handling *
 ***********************************************/
static LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
  switch (msg) {
    case WM_DISPLAYCHANGE: {
      if (!IsPaused && IsIdenticalMonitorLayout()) {
	// On some systems it's possible to disconnect and reconnect a monitor without 
	// triggering a monitor layout change, but WM_DISPLAYCHANGE is still sent.
	// Do a pause and immediate resume to catch that case
	debug((TEXT("Forced restore\n")));
	PauseWindowTracking();
	ScheduleResumeWindowTracking();
      }else if (IsDesiredMonitorLayout()) {
	// Extract monitor placements again in case the positions were changed
	DesiredMonitorPlacements.GetAllMonitors();
	ScheduleResumeWindowTracking();
      }else{
	PauseWindowTracking();
      }
    }break;
    case WM_NCDESTROY: PostQuitMessage(0); break;
  }
  return DefWindowProc(hWnd, msg, wParam, lParam);
}

static HWND CreateDetectionWnd() {
  WNDCLASS wc;
  ZeroMemory(&wc,sizeof wc);
  wc.lpfnWndProc = WindowProc;
  wc.lpszClassName = TEXT("RestoreWindows");
  RegisterClass(&wc);
  return CreateWindow(TEXT("RestoreWindows"),TEXT("RestoreWindows"),0,0,0,0,0,0,0,0,0);
}

/********************************
 * Global event listener	*
 ********************************/
static void CALLBACK WinEventProc(HWINEVENTHOOK hWinEventHook, DWORD event, HWND hwnd, LONG idObject, LONG idChild, DWORD dwEventThread, DWORD dwmsEventTime) {
  if (IsPaused) return;
  switch (event) {
    case EVENT_OBJECT_LOCATIONCHANGE:
    if (!idObject && HasWindow(hwnd)) ScheduleWindowPlacementUpdate(hwnd);
    break;

    case EVENT_OBJECT_CREATE:
    case EVENT_OBJECT_SHOW:
    if (!HasWindow(hwnd) && IsApplicationWindow(hwnd)) AddWindow(hwnd);
    break;

    case EVENT_OBJECT_DESTROY:
    case EVENT_OBJECT_HIDE:
    if (!idObject) RemoveWindow(hwnd);
    break;

    case EVENT_OBJECT_PARENTCHANGE:
    if (HasWindow(hwnd)) {
      if (!IsApplicationWindow(hwnd)) RemoveWindow(hwnd);
    }else if (IsApplicationWindow(hwnd)) AddWindow(hwnd);
    break;
  }
}

static SERVICE_STATUS_HANDLE hServiceStatus;
static SERVICE_STATUS ServiceStatus={
 SERVICE_WIN32_OWN_PROCESS,
 SERVICE_RUNNING,
 SERVICE_ACCEPT_PARAMCHANGE|SERVICE_ACCEPT_PAUSE_CONTINUE|SERVICE_ACCEPT_SHUTDOWN|SERVICE_ACCEPT_STOP
 |SERVICE_ACCEPT_HARDWAREPROFILECHANGE|SERVICE_ACCEPT_POWEREVENT|SERVICE_ACCEPT_SESSIONCHANGE,
 NO_ERROR,
 0,
 0,
 0};

static DWORD WINAPI HandlerEx(DWORD ctrl,DWORD eventtype,void*eventdata,void*context) {
  switch (ctrl) {
    case SERVICE_CONTROL_STOP: {
      DestroyWindow(DetectWindow);
    }break;
  }
  return 0;
}


static void WINAPI ServiceMain(DWORD argc, TCHAR**argv) {
  HWND w=FindWindow(TEXT("RestoreWindows"),TEXT("RestoreWindows"));
  if (w) {
    DestroyWindow(w);
    return;
  }
  hServiceStatus=RegisterServiceCtrlHandlerEx(TEXT("RestoreWindows"),HandlerEx,0);
  if (!hServiceStatus) return;

  SetServiceStatus(hServiceStatus,&ServiceStatus);
#ifdef _DEBUG
  logfile=CreateFile(TEXT("RestoreWindows.log"),GENERIC_WRITE,
   FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,0,CREATE_ALWAYS,0,0);
  if (logfile==INVALID_HANDLE_VALUE) logfile=0;
#endif
  if (argc>=1) ResumeTimerValue=clamp(0,1000*60*10,StrToInt(*argv));
  DesiredMonitorPlacements.GetAllMonitors();
  debug((TEXT("Delay: %d\n\n"), ResumeTimerValue));
  for (monvec::iterator it = DesiredMonitorPlacements.begin(); it != DesiredMonitorPlacements.end(); ++it)
		debug((TEXT("Monitor: %d, %d\n"), it->width(), it->height()));
  debug((TEXT("\n")));
  LoadAllWindowPlacements();
  HWINEVENTHOOK hWinEventHook = SetWinEventHook(EVENT_MIN, EVENT_MAX, NULL, WinEventProc, 0, 0, WINEVENT_OUTOFCONTEXT);
  DetectWindow = CreateDetectionWnd();
  MSG msg;
  while (GetMessage(&msg, NULL, 0, 0) > 0) {
    DispatchMessage(&msg);
  }
  UnhookWinEvent(hWinEventHook);
#ifdef _DEBUG
  if (logfile) CloseHandle(logfile);
#endif
}

void WINAPI WinMainCRTStartup() {
  static SERVICE_TABLE_ENTRY st[]={
   {TEXT("RestoreWindows"),ServiceMain},
   {0,0}};
  StartServiceCtrlDispatcher(st);
}
